/**
 * Copyright (c) 2000-present Liferay, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of the Liferay Enterprise
 * Subscription License ("License"). You may not use this file except in
 * compliance with the License. You can obtain a copy of the License by
 * contacting Liferay, Inc. See the License for the specific language governing
 * permissions and limitations under the License, including but not limited to
 * distribution rights of the Software.
 *
 *
 *
 */

package com.liferay.apio.architect.consumer.throwable;

import aQute.bnd.annotation.ConsumerType;

import java.util.Objects;

/**
 * Defines a {@code BiConsumer} that can throw an exception.
 *
 * <p>
 * This interface can be implemented with a lambda function.
 * </p>
 *
 * @author Alejandro Hernández
 * @param  <A> the type of the first argument of the consumer
 * @param  <B> the type of the second argument of the consumer
 */
@ConsumerType
@FunctionalInterface
public interface ThrowableBiConsumer<A, B> {

	/**
	 * Returns an empty {@code ThrowableBiConsumer} that doesn't perform any
	 * operation.
	 *
	 * @return an empty {@code ThrowableBiConsumer} that doesn't perform any
	 *         operation
	 */
	public static <A, B> ThrowableBiConsumer<A, B> empty() {
		return (a, b) -> {
		};
	}

	/**
	 * Operates with two parameters and returns {@code void}. This function can
	 * be implemented explicitly or with a lambda.
	 *
	 * @param a the first function argument
	 * @param b the second function argument
	 */
	public void accept(A a, B b) throws Exception;

	/**
	 * Returns the {@code ThrowableBiConsumer} function that first executes the
	 * current {@code ThrowableBiConsumer} instance's {@code accept} method,
	 * then executes the {@code after} parameter's {@code accept} method.
	 *
	 * @param  after the {@code ThrowableBiConsumer} instance to execute after
	 *         the current instance
	 * @return the {@code ThrowableBiConsumer} that executes the current
	 *         instance's {@code accept} method, as well as that of {@code
	 *         after}
	 */
	public default ThrowableBiConsumer<A, B> andThen(
		ThrowableBiConsumer<? super A, ? super B> after) {

		Objects.requireNonNull(after);

		return (a, b) -> {
			accept(a, b);
			after.accept(a, b);
		};
	}

}